/******************************************************************************

Copyright (c) 1997, Andrew J. Weber, Jason P. Weber

Copyright information should have been included with this distribution
and may be found at ../meta/COPYRIGHT or obtained via email from weber@acm.org

******************************************************************************/

#include <bld/environ.h>
#include <bld/fnop.h>
#include <bld/interp.h>

/*!!KLUDGE
#if OSD_OS==OSD_HPUX
extern "C" {
#endif
*/
extern int yyparse(void);
/*
#if OSD_OS==OSD_HPUX
}
#endif
*/

extern FILE *yyin;

/***********************************************************************
class BLD_Interpreter
***********************************************************************/
char *BLD_Interpreter::parseFileName = "";
GPL_Boolean BLD_Interpreter::parseSuccess = FALSE;
unsigned long BLD_Interpreter::lineNumber = 0;
OpArray *BLD_Interpreter::operations = (OpArray *) 0;
const GPL_String BLD_Interpreter::defaultTargetName("__DEFAULT__");

BLD_Interpreter::BLD_Interpreter(char **e)
	: WBA_Application(this), states(GPL_String::HashFunction, 64), envp(e)
{
}

BLD_Interpreter::~BLD_Interpreter(void)
{
}

GPL_Boolean BLD_Interpreter::Initialize(void)
{
	version = gplCoralVersion;

	// Set the default delivery.
	WBA_LOGGER.SetDefaultDelivery(new WBA_DeliveryPrintf());

	if (!WBA_Application::Initialize())
	{
		return FALSE;
	}

	if (BLD_ARG_ON("BLD_PRINT_HELP"))
	{
		Printf("%s", GetUsage());
		SetExitCode(0);
		Exit();
	}
	if (BLD_ARG_ON("BLD_PRINT_VERSION"))
	{
		Printf("%s %s", GetName(), GetVersion().ToString().GetData());
		SetExitCode(0);
		Exit();
	}

	// Stuff any command-line parameters into the targets EV.
	GPL_String targets = WBA_ARGS.Find("BLD_TARGETS")->GetValue();
	targets.Join(" ", GPL_String::Join(" ", *(WBA_ENV.GetParameters())));
	WBA_ARGS.Find("BLD_TARGETS")->SetValue(targets);

	return TRUE;
}

void BLD_Interpreter::Finalize(void)
{
	WBA_Application::Finalize();
}

void BLD_Interpreter::Process(void)
{
	// Determine the build file name.
	register WBA_Argument *file = WBA_ARGS.Find("BLD_FILE");
	if ((WBA_Argument *) 0 == file || 0 == file->GetValue().GetLength())
	{
		Printf("No build file specified");
		SetExitCode(-1);
		return;
	}

	// Make sure the build file exists.
	if (!GPL_File::Exists(file->GetValue().GetData()))
	{
		Printf("%s does not exist", file->GetValue().GetData());
		SetExitCode(-1);
		return;
	}

	// Build all of the targets.
	if (BLD_ARG_ON("BLD_DEBUG"))
	{
		WBA_APP->Printf("Using build file %s", file->GetValue().GetData());
	}
	if (!Build(file->GetValue(), WBA_ARGS.Find("BLD_TARGETS")->GetValues(" ")))
	{
		SetExitCode(-1);
	}
}

GPL_Boolean BLD_Interpreter::Parse(const GPL_String & fileName, OpArray *ops)
{
	// Check for the existence of the build file.
	if (!GPL_File::Exists(fileName))
	{
		Printf("%s does not exist", fileName.GetData());
		SetExitCode(-2);
		return FALSE;
	}

	// Parse the file.
	parseFileName = fileName.GetData();
	parseSuccess = TRUE;
	lineNumber = 1;
	operations = ops;
	yyin = fopen(parseFileName, "r");
	yyparse();
	fclose(yyin);
	if (parseSuccess)
	{
		if (BLD_ARG_ON("BLD_DEBUG"))
		{
			Printf("Parsed %s", fileName.GetData());
		}
		return TRUE;
	}
	return FALSE;
}

GPL_Boolean BLD_Interpreter::Build(const GPL_String & fileName,
									const GPL_Array<GPL_String> & names)
{
	// Set up the root context.
	BLD_Context rootContext;
	contexts.Push(&rootContext);
	BLD_Environment::GetEnvironment()->SetEnvironmentVariables(envp);
	BLD_Value targets(names);
	rootContext.SetValue("BLD_TARGETS", targets);
	BLD_Value plat(osdPlatform(NULL));
	rootContext.SetValue("BLD_PLATFORM", plat);
	BLD_Value hw(osdPlatform("h"));
	rootContext.SetValue("BLD_HARDWARE", hw);
	BLD_Value os(osdPlatform("o"));
	rootContext.SetValue("BLD_OS", os);
	BLD_Value osver(osdPlatform("r"));
	rootContext.SetValue("BLD_OSVER", osver);
	BLD_Value comp(osdPlatform("c"));
	rootContext.SetValue("BLD_COMPILER", comp);
	BLD_Value ostype(osdPlatform("s"));
	rootContext.SetValue("BLD_OSTYPE", ostype);

	// Parse the build file.
	OpArray ops;
	if (!Parse(fileName, &ops))
	{
		Printf("Unable to parse %s", fileName.GetData());
		contexts.Pop();
		return FALSE;
	}

	// Set up the default target.
	BLD_Target defaultTarget;
	defaultTarget.SetName(defaultTargetName);
	BLD_Rule defaultRule(BLD_RULE_OVERWRITE_OPERATIONS, ops);
	GPL_Array<GPL_String> expressions;
	defaultTarget.AddRule((BLD_RuleBase *) &defaultRule, expressions);

	// Build the default target, reusing the root scope.
	if (BLD_BUILD_FAILED == defaultTarget.Build(TRUE))
	{
		if (WBA_ARGS.Find("BLD_PERSIST")->IsFalse() &&
			WBA_ARGS.Find("BLD_IGNORE_TARGET_FAILURES")->IsFalse())
		{
			contexts.Pop();
			return FALSE;
		}
	}

	// Build each successive target.
	register const unsigned long n = names.GetCount();
	for (register unsigned long i = 0; n > i; ++i)
	{
		// Find the target.
		BLD_Target target;
		if (!GetCurrentContext()->CreateTarget(names[i], target))
		{
			Printf("Target %s does not exist", names[i].GetData());
			contexts.Pop();
			return FALSE;
		}

		// Build the target.
		if (BLD_ARG_ON("BLD_DEBUG"))
		{
			WBA_APP->Printf("Building target %s", names[i].GetData());
		}
		if (BLD_BUILD_FAILED == target.Build())
		{
			if (WBA_ARGS.Find("BLD_PERSIST")->IsFalse() &&
				WBA_ARGS.Find("BLD_IGNORE_TARGET_FAILURES")->IsFalse())
			{
				contexts.Pop();
				return FALSE;
			}
		}
	}
	contexts.Pop();
	return TRUE;
}

void BLD_Interpreter::AddRule (const GPL_String & string,
								const BLD_RuleType type, const OpArray & ops)
{
	if (0 < string.GetLength() && 0 < contexts.GetCount())
	{
		GetCurrentContext()->AddRule(string, type, ops);
	}
}

BLD_State & BLD_Interpreter::FindState(const GPL_String & name)
{
	// Ensure use of rooted file names.
	const GPL_String fullPathName = GPL_File::GetFullPathName(name.GetData());
//WBA_APP->Printf("%s(%s)[%s]", name.GetData(), fullPathName.GetData(), states.Find(fullPathName).ToString().GetData());
	return states.Find(fullPathName);
}

GPL_Boolean BLD_Interpreter::GetModificationTime(const GPL_String & fileName,
													GPL_Time & time)
{
	// Ensure use of rooted file names.
	const GPL_String fullPathName =
		GPL_File::GetFullPathName(fileName.GetData());
	if (GPL_File::Exists(fullPathName))
	{
		OSD_Time t;
		if (GPL_File::GetModificationTime(fullPathName, t))
		{
			time = t;
			states.Find(fullPathName).SetLastBuilt(time);
		}
		return TRUE;
	}
	return FALSE;
}

void BLD_Interpreter::DumpStates(void)
{
	GPL_MapIterator<GPL_String, BLD_State> i;
	i.Reset(&states);
	while (i.Next())
	{
		WBA_APP->Printf("%s: %s", i.Key().GetData(), i.Value().ToString().GetData());
	}
}

void BLD_Interpreter::DumpEnvironment(void)
{
	GPL_Map<GPL_String, BLD_Value> values;
	contexts.Top()->GetValues(values);

	GPL_MapIterator<GPL_String, BLD_Value> iterator;
	GPL_String value;
	iterator.Reset(&values);
	while (iterator.Next())
	{
		value = iterator.Value().GetString();
		WBA_APP->Printf("%s=%s", iterator.Key().GetData(),
						((0 == value.GetLength()) ? "" : value.GetData()));
	}
}
